SystemOrganization addCategory: #'SOUL-Repositories'!

!SoulFactory class methodsFor: '*SOUL-Repositories' stamp: 'sp 24/10/11 15:26'!
addRepositoriesDependent: aDependent
	
	RepositoriesDependents add: aDependent! !

!SoulFactory class methodsFor: '*SOUL-Repositories' stamp: 'AndyKellens 11/2/2011 09:53'!
compilingRepository
	
	| rep |
	^Repositories
		at: #compiling
		ifAbsentPut:
			[rep := SoulFactory current makeRepository new.
			rep loadLayer: QuotedParseLayer.
			QuotedParseLayer allSubclasses do: [:cl | rep loadLayer: cl].
			rep loadLayer: LogicPrimitives.
			LogicPrimitives allSubclasses do: [:cl | rep loadLayer: cl].
			self repository: #compiling put: rep.
			rep]! !

!SoulFactory class methodsFor: '*SOUL-Repositories' stamp: 'sp 24/10/11 15:26'!
doAfterRepositoryLoad: aBlock
	
	(Repositories at: #default) == nil
		ifTrue: [self addToWaitList: aBlock]
		ifFalse: [aBlock value]! !

!SoulFactory class methodsFor: '*SOUL-Repositories' stamp: 'sp 24/10/11 15:26'!
initializeRepositories
	"After loading Soul, this method will initialize the necessary repositories to ensure a correct working of the interpreter"

	
	self parsingRepository.
	self logicRepository.
	self testRepository.
	self transparentSymbiosisRepository! !

!SoulFactory class methodsFor: '*SOUL-Repositories' stamp: 'AndyKellens 11/2/2011 09:53'!
logicRepository
	
	| rep |
	rep := Repositories
		at: #logic
		ifAbsentPut:
			[rep := SoulFactory current makeRepository new.
			LogicPrimitives allSubclasses do: [:cl | rep loadLayer: cl].	"self processWaitList."

			rep].	"TO BE REMOVED IN LATER VERSIONS!!"

	Repositories at: #default put: rep.
	^rep! !

!SoulFactory class methodsFor: '*SOUL-Repositories' stamp: 'sp 24/10/11 15:26'!
notifyRepositoriesDependents
	
	RepositoriesDependents do: [:each | each updateRepositories]! !

!SoulFactory class methodsFor: '*SOUL-Repositories' stamp: 'AndyKellens 11/2/2011 09:53'!
parsingRepository
	
	| rep |
	^Repositories
		at: #parsing
		ifAbsentPut:
			[rep := SoulFactory current makeRepository new.
			rep loadLayer: QuotedParseLayer.
			QuotedParseLayer allSubclasses do: [:cl | rep loadLayer: cl].
			rep loadLayer: LogicPrimitives.
			LogicPrimitives allSubclasses do: [:cl | rep loadLayer: cl].
			self repository: #parsing put: rep.
			rep]! !

!SoulFactory class methodsFor: '*SOUL-Repositories' stamp: 'sp 24/10/11 15:26'!
removeRepositoriesDependent: aDependent
	
	RepositoriesDependents
		remove: aDependent
		ifAbsent: []! !

!SoulFactory class methodsFor: '*SOUL-Repositories' stamp: 'sp 24/10/11 15:26'!
removeRepository: aName
	
	Repositories removeKey: aName.
	self notifyRepositoriesDependents! !

!SoulFactory class methodsFor: '*SOUL-Repositories' stamp: 'sp 24/10/11 15:26'!
repository: aName
	
	aName = #logic ifTrue: [self logicRepository].	"aName = #default ifTrue:[ 	self defaultRepository ]."

	aName = #parsing ifTrue: [self parsingRepository].
	^Repositories at: aName! !

!SoulFactory class methodsFor: '*SOUL-Repositories' stamp: 'sp 24/10/11 15:26'!
repository: aName put: aRepository
	
	Repositories at: aName put: aRepository.
	aRepository name: aName.
	aRepository compile.
	self notifyRepositoriesDependents! !

!SoulFactory class methodsFor: '*SOUL-Repositories' stamp: 'sp 24/10/11 15:26'!
repositoryNames
	
	^Repositories keys! !

!SoulFactory class methodsFor: '*SOUL-Repositories' stamp: 'sp 24/10/11 15:26'!
resetDefaultRepository
	
	Repositories
		removeKey: #default
		ifAbsent: []! !

!SoulFactory class methodsFor: '*SOUL-Repositories' stamp: 'AndyKellens 11/2/2011 09:53'!
testRepository
	
	| rep |
	^Repositories
		at: #test
		ifAbsentPut:
			[rep := SoulFactory current makeRepository new.
			LogicPrimitives allSubclasses do: [:cl | rep loadLayer: cl].
			rep loadLayer: TestQueriesLayer.
			TestQueriesLayer allSubclasses do: [:cl | rep loadLayer: cl].	"self processWaitList."

			rep]! !

!SoulFactory class methodsFor: '*SOUL-Repositories' stamp: 'AndyKellens 11/2/2011 09:53'!
transparentSymbiosisRepository
	
	| rep |
	^Repositories
		at: #'transparant-symbiosis'
		ifAbsentPut:
			[rep := SoulFactory current makeRepository new.
			LogicStorage allSubclasses do: [:cl | rep loadLayer: cl].	"self processWaitList."

			rep]! !

Object subclass: #LogicStorage
	instanceVariableNames: ''
	classVariableNames: 'StorageDependents'
	poolDictionaries: ''
	category: 'SOUL-Repositories'!

LogicStorage class
	instanceVariableNames: 'selectorToPredicateKey doingInternalStorage clausesStore'!

!LogicStorage commentStamp: '<historical>' prior: 0!
The main purpose of LogicStorage and its subclasses is to serve as storage for the source of SOUL programs in a format that can be easily written to disk and especially stored on a Store facility. A Store database can only be used to hold classes and methods and can not be easily extended to allow for other kinds of data, hence we map SOUL programs onto a representation that Store understands: classes and methods. LogicStorage classes are not directly used by the evaluation to lookup rules and facts, see LogicRepository for that.

SOUL programs are organized into 'layers', each LogicStorage subclass represents one such layer.

The instance methods of a LogicStorage subclass are simply used to hold the source of the program, the class methods of LogicStorage provide a clean way of accessing and modifying the parsed representation of these programs. The parsed representation is contained in the clausesStore class instance variable.

Each instance method holds the source for all the rules and facts in the layer that implement the same predicate. The name of this method is derived from the name of the predicate. The methods are kept in two-way sync with the parsed representation, when one changes the other is updated as well.

The methods in the following protocols provide the clean interface for accessing and editing the parsed representation:

- 'accessing' provides methods to retrieve the clauses in the layer
- 'editing' provides ways for adding, removing, replacing and changing the order of the clauses in the program
- 'protocol support' provides ways of organizing the predicates contained in the layer into protocols

[ TODO: 'editing' relies on identity of clauses, err what else? ]


!

LogicStorage class
	instanceVariableNames: 'selectorToPredicateKey doingInternalStorage clausesStore'!

!LogicStorage class methodsFor: 'protocol support' stamp: 'sp 24/10/11 15:26'!
addProtocol: aSymbol
	"asSymbol just to be sure"

	
	self organization addCategory: aSymbol asSymbol! !

!LogicStorage class methodsFor: 'update mechanism' stamp: 'sp 24/10/11 15:26'!
addStorageDependentClass: aClass
	
	(self storageDependents includes: aClass)
		ifFalse: [self storageDependents add: aClass]! !

!LogicStorage class methodsFor: 'compilation support' stamp: 'sp 24/10/11 15:26'!
addSubclass: aSubclass
	
	| res |
	res := super addSubclass: aSubclass.
	self notifyLayerAdd: aSubclass.
	^res! !

!LogicStorage class methodsFor: 'accessing-all' stamp: 'sp 24/10/11 15:26'!
allClauses
	"Return all clauses in this layer and all sublayers"

	
	| collection |
	collection := self gatherClauses.
	self allSubclasses do: [:cl | collection addAll: cl gatherClauses].
	^collection! !

!LogicStorage class methodsFor: 'accessing-all' stamp: 'sp 24/10/11 15:26'!
allPredicateKeys
	"Return all predicatekeys in this layer and all sublayers"

	
	| collection |
	collection := self predicateKeys.
	self allSubclasses do: [:cl | collection addAll: cl predicateKeys].
	^collection! !

!LogicStorage class methodsFor: 'accessing-all' stamp: 'sp 24/10/11 15:26'!
allPredicateKeysIn: protocol
	"Return all predicatekeys in this layer and all sublayers"

	
	| collection |
	collection := (self predicateKeysIn: protocol) asOrderedCollection.
	self subclasses
		do: [:cl | collection addAll: (cl allPredicateKeysIn: protocol)].
	^collection! !

!LogicStorage class methodsFor: 'protocol support' stamp: 'sp 24/10/11 15:26'!
allProtocols
	
	| collection |
	collection := Set new.
	collection addAll: self protocols.
	self subclasses do: [:class | collection addAll: class allProtocols].
	^collection! !

!LogicStorage class methodsFor: 'editing-local' stamp: 'sp 24/10/11 15:26'!
assertClause: clause
	
	^self
		assertClause: clause
		inProtocol: self defaultProtocol
		ifAlreadyClassified: [:key :currentProtocol :newProtocol | false]! !

!LogicStorage class methodsFor: 'editing-local' stamp: 'sp 24/10/11 15:26'!
assertClause: clause inProtocol: protocol ifAlreadyClassified: alreadyClassifiedBlock
	
	^self
		assertClauses: (OrderedCollection with: clause)
		inProtocol: protocol
		ifAlreadyClassified: alreadyClassifiedBlock! !

!LogicStorage class methodsFor: 'editing-local' stamp: 'sp 24/10/11 15:26'!
assertClauses: clauses
	
	^self
		assertClauses: clauses
		inProtocol: self defaultProtocol
		ifAlreadyClassified: [:key :currentProtocol :newProtocol | false]! !

!LogicStorage class methodsFor: 'editing-local' stamp: 'sp 24/10/11 15:26'!
assertClauses: clauses inProtocol: protocol ifAlreadyClassified: alreadyClassifiedBlock
	
	| keys |
	keys := OrderedSet new.
	clauses
		do:
			[:clause | 
			| key |
			key := self predicateKeyofClause: clause.
			(self clausesCollectionForKey: key) add: clause.
			keys add: key].
	keys
		do:
			[:key | 
			self classifyPredicate: key under: protocol ifAlreadyClassified:
					alreadyClassifiedBlock.
			self storeClausesWithPredicateKey: key]! !

!LogicStorage class methodsFor: 'icon' stamp: 'AndyKellens 10/25/2011 10:39'!
browserIcon
	
	^nil! !

!LogicStorage class methodsFor: 'protocol support' stamp: 'sp 24/10/11 15:26'!
classifyPredicate: key under: protocol ifAlreadyClassified: alreadyClassifiedBlock
	
	| current |
	current := self protocolOfPredicate: key.
	current
		ifNil: [self movePredicateKey: key toProtocol: protocol]
		ifNotNil:
			[current = protocol
				ifFalse:
					[(alreadyClassifiedBlock value: key value: current value: protocol)
						ifTrue: [self movePredicateKey: key toProtocol: protocol]]]! !

!LogicStorage class methodsFor: 'private - storage' stamp: 'sp 24/10/11 15:26'!
clausesCollectionForKey: key
	
	^self clausesStore
		at: key
		ifAbsentPut: [OrderedCollection new]! !

!LogicStorage class methodsFor: 'accessing-all' stamp: 'sp 24/10/11 15:26'!
clausesForPredicateKey: predicateKey
	
	| clauses |
	"returns the clauses that are defined for the given predicatekey, in this layer and all sublayers"

	clauses := OrderedCollection new.
	clauses
		addAll:
			(self clausesStore
				at: predicateKey
				ifAbsent: [OrderedCollection new]).
	self subclasses
		do:
			[:subclass | clauses addAll: (subclass clausesForPredicateKey: predicateKey)].
	^clauses! !

!LogicStorage class methodsFor: 'private - storage' stamp: 'sp 24/10/11 15:26'!
clausesStore
	
	clausesStore ifNil: [self initClausesStore].
	^clausesStore! !

!LogicStorage class methodsFor: 'accessing-all' stamp: 'sp 24/10/11 15:26'!
collectPredicateNamesMatching: pattern
	
	| result |
	result := (self predicateKeys asOrderedCollection
		select: [:namex | pattern match: namex])
		collect: [:namex | Association key: self value: namex].
	self subclasses
		do:
			[:subclass | result addAll: (subclass collectPredicateNamesMatching: pattern)].
	^result! !

!LogicStorage class methodsFor: 'compilation support' stamp: 'sp 24/10/11 15:26'!
compilerClass
	
	^SoulMethodCompiler! !

!LogicStorage class methodsFor: 'accessing-local' stamp: 'sp 24/10/11 15:26'!
containsClause: clause
	
	^self gatherClauses includes: clause! !

!LogicStorage class methodsFor: 'storage-conversion' stamp: 'sp 24/10/11 15:26'!
copyFromLayerWithOldStorageSystem: oldLayer
	"This method can be used to convert layers that are still stored using one
	of the old storage systems or that were edited directly using the class
	browser rather than the clause browser. It is best to make a new layer first,
	and then send it this message using the old layer as the argument."

	
	| clauses |
	oldLayer organization categories
		do:
			[:eachCategory | 
			(oldLayer organization listAtCategoryNamed: eachCategory)
				do:
					[:selector | 
					clauses := oldLayer new perform: selector.
					self
						assertClauses: clauses
						inProtocol: eachCategory
						ifAlreadyClassified: [:predicate :oldProtocol :newProtocol | false]]]! !

!LogicStorage class methodsFor: 'repositories' stamp: 'AndyKellens 11/2/2011 09:53'!
createAnonymousRepository
	
	| rep |
	rep := SoulFactory current makeRepository new loadLayer: self.
	self allSubclasses do: [:cl | rep loadLayer: cl].
	^rep! !

!LogicStorage class methodsFor: 'private - storage' stamp: 'sp 24/10/11 15:26'!
defaultProtocol
	
	^ClassOrganizer defaultProtocol! !

!LogicStorage class methodsFor: 'private - storage' stamp: 'sp 24/10/11 15:26'!
deleteStorageOfClausesWithPredicateKey: key
	
	| selector |
	selector := self selectorForPredicateKey: key.
	self removeSelector: selector! !

!LogicStorage class methodsFor: 'private - storage' stamp: 'sp 24/10/11 15:26'!
doingInternalStorage
	
	^doingInternalStorage
		ifNil: [false]
		ifNotNil: [doingInternalStorage]! !

!LogicStorage class methodsFor: 'private - storage' stamp: 'sp 24/10/11 15:26'!
doingInternalStorage: newSettingBoolean
	
	doingInternalStorage := newSettingBoolean! !

!LogicStorage class methodsFor: 'signal constants' stamp: 'sp 24/10/11 15:26'!
errorSignal
	
	^SoulError! !

!LogicStorage class methodsFor: 'private - storage' stamp: 'sp 24/10/11 15:26'!
gatherClauses
	
	| result |
	result := OrderedCollection new.
	self clausesStore values do: [:clauses | result addAll: clauses].
	^result! !

!LogicStorage class methodsFor: 'editing-all' stamp: 'sp 24/10/11 15:26'!
inAllRename: originalName multiplicity: multiplicity to: newName argumentMapping: argumentMapping
	
	self rename: originalName multiplicity: multiplicity to: newName
		argumentMapping: argumentMapping.
	self allSubclasses
		do:
			[:layer | 
			layer rename: originalName multiplicity: multiplicity to: newName
				argumentMapping: argumentMapping]! !

!LogicStorage class methodsFor: 'initialize' stamp: 'sp 24/10/11 15:26'!
initAllPredicateMappings
	"recursively construct the table of selector -> predicates mappings. this is used by the clause browser"

	
	self initPredicateMappings.
	self subclasses do: [:subclass | subclass initAllPredicateMappings]! !

!LogicStorage class methodsFor: 'initialize' stamp: 'sp 24/10/11 15:26'!
initClausesStore
	
	clausesStore := Dictionary new! !

!LogicStorage class methodsFor: 'initialize' stamp: 'sp 24/10/11 15:26'!
initPredicateMappings
	
	| clause |
	"construct the table of selector -> predicates mappings. this is used by the clause browser"

	selectorToPredicateKey := Dictionary new.
	self selectors
		do:
			[:sel | 
			clause := (self new perform: sel) clauses first.
			selectorToPredicateKey
				at: sel
				put: (self predicateKeyofClause: clause)]! !

!LogicStorage class methodsFor: 'initialize' stamp: 'sp 24/10/11 15:26'!
initialize
	
	self initPredicateMappings.
	self initClausesStore! !

!LogicStorage class methodsFor: 'accessing-all' stamp: 'sp 24/10/11 15:26'!
layerContainingClause: clause ifNone: noneBlock
	
	| coll |
	coll := OrderedCollection with: self.
	coll addAll: self allSubclasses.
	^coll
		detect: [:layer | layer containsClause: clause]
		ifNone: noneBlock! !

!LogicStorage class methodsFor: 'private - predicate keys' stamp: 'sp 24/10/11 15:26'!
makePredicateKey: aPredicate multiplicity: aMult
	
	^aMult < 0
		ifTrue: [aPredicate , '@']
		ifFalse: [aPredicate , '/' , aMult asString]! !

!LogicStorage class methodsFor: 'editing-local' stamp: 'sp 24/10/11 15:26'!
moveClauseDown: clause
	
	| key index |
	key := self predicateKeyofClause: clause.
	index := (self clausesCollectionForKey: key)
		indexOf: clause
		ifAbsent:
			[self error: 'Trying to move a clause in a layer that does not contain it.'].
	(self clausesCollectionForKey: key) size > index
		ifTrue:
			[(self clausesCollectionForKey: key)
				swap: index
				with: index + 1.
			self storeClausesWithPredicateKey: key]! !

!LogicStorage class methodsFor: 'editing-local' stamp: 'sp 24/10/11 15:26'!
moveClauseUp: clause
	
	| key index |
	key := self predicateKeyofClause: clause.
	index := (self clausesCollectionForKey: key)
		indexOf: clause
		ifAbsent:
			[self error: 'Trying to move a clause in a layer that does not contain it.'].
	index > 1
		ifTrue:
			[(self clausesCollectionForKey: key)
				swap: index
				with: index - 1.
			self storeClausesWithPredicateKey: key]! !

!LogicStorage class methodsFor: 'protocol support' stamp: 'sp 24/10/11 15:26'!
movePredicateKey: key toProtocol: protocol
	"returns true if successful, to see if we need to refresh any views"

	
	| selector org oldProtocol |
	selector := self selectorForPredicateKey: key.
	org := self organization.	"The below code is cleaner, but Store does not notice anything has changed when we do this."
	"TODO: Make the store notice!!"

	org := self organization.
	oldProtocol := org categoryOfElement: selector.
	oldProtocol = protocol
		ifFalse:
			[org removeElement: selector.
			org classify: selector under: protocol.
			^true].
	^false! !

!LogicStorage class methodsFor: 'private - predicate keys' stamp: 'sp 24/10/11 15:26'!
multiplicityOfKey: aPredicateKey
	
	| splitIndex multString |
	splitIndex := aPredicateKey indexOf: $/.
	multString := aPredicateKey
		copyFrom: splitIndex + 1
		to: aPredicateKey size.
	^multString asNumber! !

!LogicStorage class methodsFor: 'update mechanism' stamp: 'sp 24/10/11 15:26'!
notifyImplementationChange
	
	self storageDependents
		do:
			[:class | 
			class withAllSubclasses
				do:
					[:aclass | aclass allInstancesDo: [:e | e update: #changePredicate with: self]]]! !

!LogicStorage class methodsFor: 'update mechanism' stamp: 'sp 24/10/11 15:26'!
notifyLayerAdd: aClass
	
	self storageDependents
		do:
			[:class | class allInstances do: [:e | e update: #addLayer with: aClass]]! !

!LogicStorage class methodsFor: 'update mechanism' stamp: 'sp 24/10/11 15:26'!
notifyLayerRemove: aClass
	
	self storageDependents
		do:
			[:class | class allInstances do: [:e | e update: #removeLayer with: aClass]]! !

!LogicStorage class methodsFor: 'private - storage' stamp: 'AndyKellens 11/2/2011 09:53'!
parseInstanceStorageMethod: selector
	
	^SoulFactory current makeParser
		parseAfterRemovingSelector: (self sourceCodeAt: selector)! !

!LogicStorage class methodsFor: 'private - predicate keys' stamp: 'sp 24/10/11 15:26'!
predicateKeyofClause: aClause
	
	^self
		makePredicateKey: aClause predicate
		multiplicity: aClause multiplicity! !

!LogicStorage class methodsFor: 'accessing-local' stamp: 'sp 24/10/11 15:26'!
predicateKeys
	"return all predicatekeys in this layer"

	
	^self selectorToPredicateKeyMapping values! !

!LogicStorage class methodsFor: 'accessing-local' stamp: 'sp 24/10/11 15:26'!
predicateKeysIn: protocol
	
	| mapping |
	mapping := self selectorToPredicateKeyMapping.
	^(self organization listAtCategoryNamed: protocol)
		collect: [:selector | mapping at: selector]! !

!LogicStorage class methodsFor: 'test classes' stamp: 'sp 24/10/11 15:26'!
predicateKeysWithoutTestCases
	
	| missingTestKeys |
	missingTestKeys := OrderedCollection new.
	self predicateKeys
		do:
			[:key | 
			self
				testMethodForPredicateKey: key
				ifAbsent: [missingTestKeys add: key]].
	^missingTestKeys! !

!LogicStorage class methodsFor: 'private - predicate keys' stamp: 'sp 24/10/11 15:26'!
predicateOfKey: aPredicateKey
	
	^aPredicateKey copyUpTo: $/! !

!LogicStorage class methodsFor: 'private - storage' stamp: 'sp 24/10/11 15:26'!
printClausesWithKey: key
	
	| printStream |
	printStream := WriteStream on: String new.
	printStream
		cr;
		cr.
	(self clausesCollectionForKey: key)
		do: [:cl | cl printOn: printStream]
		separatedBy:
			[printStream
				write: '.';
				cr;
				cr].
	^printStream contents! !

!LogicStorage class methodsFor: 'protocol support' stamp: 'sp 24/10/11 15:26'!
protocolOfPredicate: aString
	
	| selector org |
	selector := self selectorForPredicateKey: aString.
	org := self organization.
	^org categoryOfElement: selector! !

!LogicStorage class methodsFor: 'protocol support' stamp: 'sp 24/10/11 15:26'!
protocols
	
	^self organization categories! !

!LogicStorage class methodsFor: 'private - storage' stamp: 'sp 24/10/11 15:26'!
reReadAllStorage
	
	self reReadFromStorage.
	self subclasses do: [:class | class reReadAllStorage]! !

!LogicStorage class methodsFor: 'private - storage' stamp: 'sp 24/10/11 15:26'!
reReadFromStorage
	
	| dummyInstance clauses key |
	dummyInstance := self new.
	self initClausesStore.
	self initPredicateMappings.
	self selectors
		do:
			[:selector | 
			clauses := (dummyInstance perform: selector) clauses.
			key := self predicateKeyofClause: clauses first.
			self clausesStore at: key put: clauses.
			self selectorToPredicateKeyMapping at: selector put: key].
	self notifyImplementationChange! !

!LogicStorage class methodsFor: 'compilation support' stamp: 'sp 24/10/11 15:26'!
recompileAll
	
	self recompileSelectors.
	self subclasses do: [:subclass | subclass recompileAll]! !

!LogicStorage class methodsFor: 'compilation support' stamp: 'sp 24/10/11 15:26'!
recompileSelectors
	
	self selectors do: [:selector | self recompile: selector]! !

!LogicStorage class methodsFor: 'protocol support' stamp: 'sp 24/10/11 15:26'!
removeProtocol: protocolName
	
	(self predicateKeysIn: protocolName asSymbol)
		do: [:predicateKey | self retractClausesWithPredicateKey: predicateKey].
	self organization removeCategory: protocolName asSymbol! !

!LogicStorage class methodsFor: 'compilation support' stamp: 'sp 24/10/11 15:26'!
removeSelector: sel
	
	| res |
	self clausesStore removeKey: (selectorToPredicateKey at: sel).
	selectorToPredicateKey
		removeKey: sel
		ifAbsent: [].
	res := super removeSelector: sel.
	self notifyImplementationChange.
	^res! !

!LogicStorage class methodsFor: 'update mechanism' stamp: 'sp 24/10/11 15:26'!
removeStorageDependentClass: aClass
	
	self storageDependents remove: aClass! !

!LogicStorage class methodsFor: 'compilation support' stamp: 'sp 24/10/11 15:26'!
removeSubclass: aSubclass
	
	| res |
	res := super removeSubclass: aSubclass.
	self notifyLayerRemove: aSubclass.
	^res! !

!LogicStorage class methodsFor: 'editing-local' stamp: 'sp 24/10/11 15:26'!
rename: renamed multiplicity: multiplicity to: newName argumentMapping: argumentMapping
	
	| renameVisitor newClause |
	renameVisitor := CompoundTermRenamingVisitor rename: renamed multiplicity:
			multiplicity to: newName argumentMapping: argumentMapping.
	self gatherClauses
		do:
			[:clause | 
			newClause := clause accept: renameVisitor.
			renameVisitor didRenaming
				ifTrue: [self replaceClause: clause with: newClause].
			renameVisitor resetDidRenaming]! !

!LogicStorage class methodsFor: 'protocol support' stamp: 'sp 24/10/11 15:26'!
renameProtocol: oldName to: newName
	"asSymbol just to be sure"

	
	self organization
		renameCategory: oldName asSymbol
		to: newName asSymbol! !

!LogicStorage class methodsFor: 'editing-local' stamp: 'sp 24/10/11 15:26'!
replaceClause: original with: replacement
	
	^self
		replaceClause: original
		with: replacement
		inProtocol:
			(self protocolOfPredicate: (self predicateKeyofClause: original))! !

!LogicStorage class methodsFor: 'editing-local' stamp: 'sp 24/10/11 15:26'!
replaceClause: original with: replacement inProtocol: protocol
	
	^self
		replaceClause: original
		with: replacement
		inProtocol: protocol
		ifAlreadyClassified: [:key :currentProtocol :newProtocol | false]! !

!LogicStorage class methodsFor: 'editing-local' stamp: 'sp 24/10/11 15:26'!
replaceClause: original with: replacement inProtocol: protocol ifAlreadyClassified: alreadyClassifiedBlock
	
	| keyOriginal keyReplacement indexOriginal |
	keyOriginal := self predicateKeyofClause: original.
	keyReplacement := self predicateKeyofClause: replacement.
	keyOriginal = keyReplacement
		ifTrue:
			[indexOriginal := (self clausesCollectionForKey: keyOriginal)
				indexOf: original
				ifAbsent:
					[self error:
							'Trying to replace a clause in a layer which does not contain it.'].
			(self clausesCollectionForKey: keyOriginal) at: indexOriginal put:
					replacement.
			self storeClausesWithPredicateKey: keyOriginal]
		ifFalse:
			[self retractClause: original.
			self assertClause: replacement inProtocol: protocol ifAlreadyClassified:
					alreadyClassifiedBlock]! !

!LogicStorage class methodsFor: 'editing-local' stamp: 'sp 24/10/11 15:26'!
retractClause: clause
	
	| key |
	key := self predicateKeyofClause: clause.
	(self clausesCollectionForKey: key)
		remove: clause
		ifAbsent:
			[self error:
					'Trying to remove a clause from a layer that does not contain it.'].
	self storeClausesWithPredicateKey: key! !

!LogicStorage class methodsFor: 'editing-local' stamp: 'sp 24/10/11 15:26'!
retractClausesWithPredicateKey: key
	
	self deleteStorageOfClausesWithPredicateKey: key! !

!LogicStorage class methodsFor: 'private - storage' stamp: 'sp 24/10/11 15:26'!
selectorForPredicateKey: aKey
	
	^((((aKey copyReplaceAll: ':' with: 'X') copyReplaceAll: '/' with: 'M')
		copyReplaceAll: '=' with: 'E') copyReplaceAll: '@' with: 'Q') asSymbol! !

!LogicStorage class methodsFor: 'private - storage' stamp: 'sp 24/10/11 15:26'!
selectorToPredicateKeyMapping
	
	selectorToPredicateKey == nil ifTrue: [self initPredicateMappings].
	^selectorToPredicateKey! !

!LogicStorage class methodsFor: 'private - storage' stamp: 'AndyKellens 11/2/2011 09:53'!
storageChangedExternally: selector newCode: aCode
	
	| clauses key |
	clauses := (SoulFactory current makeParser parseAfterRemovingSelector: aCode)
		clauses.
	key := self predicateKeyofClause: clauses first.
	self clausesStore at: key put: clauses.
	self selectorToPredicateKeyMapping at: selector put: key.
	self notifyImplementationChange! !

!LogicStorage class methodsFor: 'private - storage' stamp: 'sp 24/10/11 15:26'!
storageDependents
	
	StorageDependents ifNil: [StorageDependents := OrderedCollection new].
	^StorageDependents! !

!LogicStorage class methodsFor: 'private - storage' stamp: 'sp 24/10/11 15:26'!
storeClausesWithPredicateKey: key
	
	| source selector |
	self doingInternalStorage: true.
	[(self clausesCollectionForKey: key) isEmpty
		ifTrue: [self deleteStorageOfClausesWithPredicateKey: key]
		ifFalse:
			[selector := self selectorForPredicateKey: key.
			source := selector , (self printClausesWithKey: key).
			self compile: source.
			self selectorToPredicateKeyMapping at: selector put: key.
			self notifyImplementationChange]]
		ensure: [self doingInternalStorage: false]! !

!LogicStorage class methodsFor: 'editing-local' stamp: 'sp 24/10/11 15:26'!
swapClause: clauseA with: clauseB
	
	| key indexA indexB |
	key := self predicateKeyofClause: clauseA.
	(self predicateKeyofClause: clauseB) = key
		ifFalse:
			[self error:
					'Trying to swap two clauses which do not implement the same predicate.'].
	indexA := (self clausesCollectionForKey: key)
		indexOf: clauseA
		ifAbsent:
			[self error: 'Trying to swap two clauses not in the same layer. '].
	indexB := (self clausesCollectionForKey: key)
		indexOf: clauseB
		ifAbsent:
			[self error: 'Trying to swap two clauses not in the same layer. '].
	(self clausesCollectionForKey: key) swap: indexA with: indexB.
	self storeClausesWithPredicateKey: key! !

!LogicStorage class methodsFor: 'test classes' stamp: 'sp 24/10/11 15:26'!
testClassIfAbsent: absentBlock
	
	^self
		testClassNamed: self name , 'Tests'
		ifAbsent: absentBlock! !

!LogicStorage class methodsFor: 'test classes' stamp: 'sp 24/10/11 15:26'!
testClassNamed: testClassName ifAbsent: absentBlock
	
	SoulTests at: testClassName ifAbsent: absentBlock! !

!LogicStorage class methodsFor: 'test classes' stamp: 'sp 24/10/11 15:26'!
testMethodForPredicateKey: key ifAbsent: absentBlock
	
	| testSelector testClass |
	testSelector := 'test'
		, ((key copyReplaceAll: '/' with: '') copyReplaceAll: ':' with: '_').
	testClass := self testClassIfAbsent: [^absentBlock value].
	^testClass methodDictionary at: testSelector ifAbsent: absentBlock! !

!LogicStorage class methodsFor: 'private' stamp: 'sp 24/10/11 15:26'!
usersOfPredicate: pred multiplicity: mult
	
	^self gatherClauses
		select: [:clause | clause usesPredicate: pred multiplicity: mult]! !

!LogicStorage class methodsFor: 'accessing-all' stamp: 'sp 24/10/11 15:26'!
usersOfPredicateKey: aPredicateKey
	
	| pred mult result |
	pred := self predicateOfKey: aPredicateKey.
	mult := self multiplicityOfKey: aPredicateKey.
	result := self usersOfPredicate: pred multiplicity: mult.
	self allSubclasses
		do:
			[:layer | result addAll: (layer usersOfPredicate: pred multiplicity: mult)].
	^result! !

SimpleTermVisitor subclass: #LexicalAddressVisitor
	instanceVariableNames: 'database'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SOUL-Repositories'!

!LexicalAddressVisitor class methodsFor: 'instance creation' stamp: 'sp 24/10/11 15:26'!
on: aDatabase
	
	^super new database: aDatabase! !

!LexicalAddressVisitor methodsFor: 'visiting terms' stamp: 'sp 24/10/11 15:26'!
callTermVisit: aCallTerm
	"do nothing... they cannot be lexicalized"
	
	! !

!LexicalAddressVisitor methodsFor: 'visiting terms' stamp: 'sp 24/10/11 15:26'!
compoundVisit: aCompound
	
	| number |
	number := database addressForPredicate: aCompound predicate.
	aCompound lexicalAddress: number.
	aCompound functor accept: self.
	aCompound termSequence accept: self.
	aCompound astLink == nil ifFalse: [aCompound astLink accept: self]! !

!LexicalAddressVisitor methodsFor: 'visiting terms' stamp: 'sp 24/10/11 15:26'!
constantVisit: aConstant
	
	| number |
	number := database addressForPredicate: aConstant predicate.
	aConstant lexicalAddress: number! !

!LexicalAddressVisitor methodsFor: 'initialize' stamp: 'sp 24/10/11 15:26'!
database: aDatabase
	
	database := aDatabase! !

!LexicalAddressVisitor methodsFor: 'visiting terms' stamp: 'sp 24/10/11 15:26'!
keywordFunctorVisit: term
	"For convenience this method will by default forward the message request to multiPartFunctorVisit: as the visitor's behavior for keyword en message functors will in most cases be the same"

	
	^self multiPartFunctorVisit: term! !

!LexicalAddressVisitor methodsFor: 'visiting terms' stamp: 'sp 24/10/11 15:26'!
messageFunctorVisit: term
	"For convenience this method will by default forward the message request to multiPartFunctorVisit: as the visitor's behavior for keyword en message functors will in most cases be the same"

	
	^self multiPartFunctorVisit: term! !

!LexicalAddressVisitor methodsFor: 'visiting terms' stamp: 'sp 24/10/11 15:26'!
multiPartFunctorVisit: aFunctor
	"See comment in keywordFunctorVisit: and messageFunctorVisit:"

	
	^aFunctor keywords do: [:keyword | keyword accept: self]! !

!LexicalAddressVisitor methodsFor: 'visiting terms' stamp: 'sp 24/10/11 15:26'!
objectVisit: anObject
	
	^self! !

!LexicalAddressVisitor methodsFor: 'visiting terms' stamp: 'sp 24/10/11 15:26'!
resultMessageTermVisit: rmTerm
	
	rmTerm message accept: self.
	rmTerm result accept: self! !

!LexicalAddressVisitor methodsFor: 'visiting terms' stamp: 'sp 24/10/11 15:26'!
smalltalkTermVisit: anSTTerm
	"TODO: should we also recursively visit the variables used in the smalltalk term by default?"
	
	! !

!LexicalAddressVisitor methodsFor: 'visiting terms' stamp: 'sp 24/10/11 15:26'!
unaryMessageFunctorVisit: unaryMessageFunctor
	
	^unaryMessageFunctor message accept: self! !

SimpleTermVisitor subclass: #RepositoryVariableAccVisitor
	instanceVariableNames: 'variables'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SOUL-Repositories'!

!RepositoryVariableAccVisitor class methodsFor: 'instance creation' stamp: 'sp 24/10/11 15:26'!
new
	
	^super new initialize! !

!RepositoryVariableAccVisitor methodsFor: 'visiting terms' stamp: 'sp 24/10/11 15:26'!
callTermVisit: aCallTerm
	
	variables add: aCallTerm variable! !

!RepositoryVariableAccVisitor methodsFor: 'initialize' stamp: 'sp 24/10/11 15:26'!
initialize
	
	variables := OrderedCollection new! !

!RepositoryVariableAccVisitor methodsFor: 'access' stamp: 'sp 24/10/11 15:26'!
variables
	
	^variables! !

Model subclass: #LogicRepository
	instanceVariableNames: 'clauseArray addressForPredicate currentAddress maxAddress fromLayers repositoryVariables exports storageChanged name compilationFactory tabledRules callGraph additionalClauses'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SOUL-Repositories'!

!LogicRepository commentStamp: '<historical>' prior: 0!
LogicRepository represents a database of logic clauses that can be used by the Soul interpreter. It supports adding clauses from layers as well as adding new clauses that are not saved in a layer. 

Normally, the interpreter fetches the clauses by lexical address. This can be done because a repository can 'compile' his clauses to make repository access faster. Compilation of the repository simply means that all predicates in the repository are given a unique address, at which they can be found in the repository. The interpreter can thus easily and quickly access all necessary clauses.

Repositories can also be composed through repository variables. These are variables that can be used in Soul's callterms (e.g. ?var->pred(a) ). 
TODO: we must find a way to unify these variables with normal variables occuring in the rest of the program.

!

!LogicRepository class methodsFor: 'signal constants' stamp: 'sp 24/10/11 15:26'!
errorSignal
	
	^SoulRuntimeException! !

!LogicRepository class methodsFor: 'initialize' stamp: 'sp 24/10/11 15:26'!
initialize
	
	LogicStorage addStorageDependentClass: self! !

!LogicRepository class methodsFor: 'instance creation' stamp: 'sp 24/10/11 15:26'!
new
	"Create an empty logic repository"

	
	^super new initialize! !

!LogicRepository methodsFor: 'add/del - runtime' stamp: 'sp 24/10/11 15:26'!
addCallClause: aClause for: aVariable
	
	| coll |
	"callback from the clause to add a clause with a callterm as conclusion to the repository"

	aClause makeClean.
	aClause definitionRepository: self.
	coll := exports
		at: aVariable name
		ifAbsentPut: [OrderedCollection new].
	coll add: aClause! !

!LogicRepository methodsFor: 'add/del - runtime' stamp: 'sp 24/10/11 15:26'!
addClause: aClause
	"Add a clause to this repository"

	
	aClause addToRepository: self.
	self addClauses: aClause additionalClausesForDefinitionRepository.
	^true! !

!LogicRepository methodsFor: 'add/del - from storage' stamp: 'sp 24/10/11 15:26'!
addClauses: aCollection
	
	aCollection do: [:aClause | self addClause: aClause]	"The following code assumes that all clauses have the same predicate _and_ multiplicity

	| address result multiplicity pred |
	pred := aCollection first predicate.
	multiplicity := aCollection first multiplicity.
	pred
	address := self storeAddressForPredicate: pred.
	aCollection do:[:clause | clause makeClean. clause definitionRepository: self].
	result := self clausesAt: address multiplicity: multiplicity.
	result isEmpty
		ifTrue: [self clausesAt: address multiplicity: multiplicity put: aCollection]
		ifFalse: [self clausesAt: address multiplicity: multiplicity put: result , aCollection]"! !

!LogicRepository methodsFor: 'add/del - from storage' stamp: 'sp 24/10/11 15:26'!
addLayers
	
	| instance program |
	fromLayers
		do:
			[:aLayer | 
			instance := aLayer new.
			aLayer selectors
				do:
					[:sel | 
					program := instance perform: sel.
					self addClauses: program clauses]]! !

!LogicRepository methodsFor: 'add/del - runtime' stamp: 'sp 24/10/11 15:26'!
addNormalClause: aClause
	
	| pred multi address aCollection |
	"callback from the clause to add a normal clause to the repository"

	pred := aClause predicate.
	multi := aClause multiplicity.
	address := self storeAddressForPredicate: pred.
	aCollection := self clausesAt: address multiplicity: multi.
	aClause makeClean.
	aClause definitionRepository: self.
	aCollection add: aClause.
	self clausesAt: address multiplicity: multi put: aCollection! !

!LogicRepository methodsFor: 'add/del - runtime' stamp: 'sp 24/10/11 15:26'!
addProgram: aProgram
	
	| clauses |
	"Add a Soul program to the database"

	clauses := aProgram clauses.
	clauses do: [:cl | self addClause: cl]! !

!LogicRepository methodsFor: 'add/del - from storage' stamp: 'sp 24/10/11 15:26'!
addTabledRule: aTabledRule
	
	tabledRules add: aTabledRule! !

!LogicRepository methodsFor: 'prepare for use' stamp: 'sp 24/10/11 15:26'!
additionalClausesToBeAdded: aCollection
	"From an evaluated query..."

	
	additionalClauses := aCollection! !

!LogicRepository methodsFor: 'access' stamp: 'sp 24/10/11 15:26'!
addressForPredicate: predicate
	"returns the lexical address assigned to a predicate. returns nil if no address was assigned to that predicate"

	
	^addressForPredicate
		at: predicate
		ifAbsent: [nil]! !

!LogicRepository methodsFor: 'private' stamp: 'sp 24/10/11 15:26'!
allClausesAt: anAddress
	
	| multArray resultClauses clauses |
	"return all clauses at the address"

	multArray := clauseArray at: anAddress.
	resultClauses := OrderedCollection new.
	multArray == nil ifTrue: [^resultClauses].
	(1 to: multArray size)
		do:
			[:index | 
			clauses := multArray at: index.
			clauses == nil ifFalse: [resultClauses addAll: clauses]].
	^resultClauses! !

!LogicRepository methodsFor: 'private' stamp: 'sp 24/10/11 15:26'!
allClausesWith: aMultiplicity
	"return all clauses with a specific multiplicity"

	
	| resultClauses clauses |
	resultClauses := OrderedCollection new.
	clauseArray
		do:
			[:multArray | 
			multArray == nil
				ifFalse:
					[clauses := multArray at: aMultiplicity.
					clauses == nil ifFalse: [resultClauses addAll: clauses]]].
	^resultClauses! !

!LogicRepository methodsFor: 'add/del - runtime' stamp: 'AndyKellens 11/2/2011 09:53'!
assert: aText
	
	| program |
	"Add a program in textformat to the repository"

	program := SoulFactory current makeParser
		parse: aText
		onError: [:e :i | self error: 'Not a valid Soul clause'].
	self addProgram: program.
	^true! !

!LogicRepository methodsFor: 'repository composition' stamp: 'sp 24/10/11 15:26'!
bind: aVariable with: aRepository
	
	| varName |
	"Bind a repository variable to a repository"

	varName := aVariable name.
	self bindName: varName with: aRepository! !

!LogicRepository methodsFor: 'repository composition' stamp: 'sp 24/10/11 15:26'!
bindName: varName with: aRepository
	"Bind a repository variable to a repository"

	
	repositoryVariables at: varName put: aRepository.
	exports
		at: varName
		ifAbsent: [^true].
	(exports at: varName) do: [:cl | aRepository importClause: cl]! !

!LogicRepository methodsFor: 'repository composition' stamp: 'sp 24/10/11 15:26'!
breakBindingFor: aVariable
	"Free a repository variable"

	
	^self breakBindingForName: aVariable name! !

!LogicRepository methodsFor: 'repository composition' stamp: 'sp 24/10/11 15:26'!
breakBindingForName: varName
	
	| repository |
	"Free a repository variable"

	repository := repositoryVariables
		at: varName
		ifAbsent: [self error: varName , ' does not exist!!'].
	repositoryVariables removeKey: varName.
	exports
		at: varName
		ifAbsent: [^true].
	(exports at: varName) do: [:cl | repository removeClause: cl]! !

!LogicRepository methodsFor: 'private' stamp: 'sp 24/10/11 15:26'!
clausesAt: anAddress multiplicity: aMultiplicity
	
	| multIndex multArray |
	"return all clauses at the address for the given multiplicity"

	multIndex := aMultiplicity + 2.
	multArray := clauseArray at: anAddress.
	multArray == nil ifTrue: [^OrderedCollection new].
	^multIndex > multArray size
		ifTrue: [OrderedCollection new]
		ifFalse: [multArray at: multIndex]! !

!LogicRepository methodsFor: 'private' stamp: 'sp 24/10/11 15:26'!
clausesAt: anAddress multiplicity: aMultiplicity put: aCollection
	
	| multIndex multArray |
	"set a collection of clauses at the address for the given multiplicity"

	multIndex := aMultiplicity + 2.
	multArray := clauseArray at: anAddress.
	multArray == nil
		ifTrue:
			[multArray := Array new: 20.
			(1 to: multArray size)
				do:
					[:i | 
					multArray
						at: i
						put: OrderedCollection new].
			clauseArray at: anAddress put: multArray].	"TODO: allow the multiplicity to go over 18"

	^multIndex > multArray size
		ifTrue:
			[self halt:
					'This has not been implemented yet. Try making shorter predicates']
		ifFalse: [multArray at: multIndex put: aCollection]! !

!LogicRepository methodsFor: 'access' stamp: 'sp 24/10/11 15:26'!
clausesDo: aBlock
	
	| multArray clauses |
	self compile.	"clauseArray do: [ :multArray |
		multArray ifNotNil: [ x do: [ :clauseCollection | 
			clauseCollection ifNotNil: [ clauseCollection do: [ :clause |
				aBlock value: clause
			] ]
		] ]
	]"

	(1 to: currentAddress)
		do:
			[:index | 
			multArray := clauseArray at: index.
			(1 to: multArray size)
				do:
					[:idx | 
					clauses := multArray at: idx.
					clauses do: [:clause | aBlock value: clause]]]! !

!LogicRepository methodsFor: 'private' stamp: 'AndyKellens 11/2/2011 09:53'!
clear
	
	SoulFactory resetSLGTables.
	tabledRules := OrderedCollection new.
	clauseArray := Array new: maxAddress.
	addressForPredicate := Dictionary new.
	currentAddress := 0.	" We should do something with these
	repositoryVariables := Dictionary new."

	exports := Dictionary new.
	self loadNatives! !

!LogicRepository methodsFor: 'access' stamp: 'AndyKellens 10/28/2011 16:09'!
compilationFactory
	^compilationFactory! !

!LogicRepository methodsFor: 'access' stamp: 'AndyKellens 10/28/2011 16:09'!
compilationFactory: aFactory

	compilationFactory := aFactory! !

!LogicRepository methodsFor: 'prepare for use' stamp: 'AndyKellens 10/28/2011 16:04'!
compile
	"Compile the clauses in this repository if the changed flag indicates that it is no longer in sync with the storage"
	^self compileForced: additionalClauses isEmpty not! !

!LogicRepository methodsFor: 'prepare for use' stamp: 'AndyKellens 11/2/2011 09:53'!
compileForced: aBoolean
	"Compile the clauses in this repository if the changed flag indicates that it is no longer in sync with the storage"

	
	| time |
	(aBoolean
		or: [storageChanged or: [(compilationFactory = SoulFactory current) not]])
		ifTrue:
			[time := Time
				millisecondsToRun:
					[self clear.
					self addLayers.
					self addClauses: additionalClauses.	"self createCallGraph."

					self transformTabledRules.
					self createLexicalAddresses.
					storageChanged := false.
					compilationFactory := SoulFactory current].
			Transcript
				show: 'Recompiled repository in ' , time asString , ' ms.';
				cr]! !

!LogicRepository methodsFor: 'private' stamp: 'sp 24/10/11 15:26'!
createCallGraph
	
	| clauses |
	clauses := OrderedCollection new.
	clauseArray
		do:
			[:multArray | 
			multArray
				ifNotNil:
					[multArray
						do:
							[:clauseCollection | 
							clauseCollection
								ifNotNil: [clauseCollection do: [:clause | clauses add: clause]]]]].
	callGraph := CallGraph withClauses: clauses! !

!LogicRepository methodsFor: 'private' stamp: 'sp 24/10/11 15:26'!
createLexicalAddresses
	"set lexicaladdresses on all clauses in the repository"

	
	| visitor multArray clauses |
	visitor := LexicalAddressVisitor on: self.
	(1 to: currentAddress)
		do:
			[:index | 
			multArray := clauseArray at: index.
			(1 to: multArray size)
				do:
					[:idx | 
					clauses := multArray at: idx.
					clauses do: [:clause | clause accept: visitor]]]! !

!LogicRepository methodsFor: 'private' stamp: 'sp 24/10/11 15:26'!
expandStorage: aSize
	
	| newArray |
	"grow the repository"

	maxAddress := maxAddress + aSize.
	newArray := Array new: maxAddress.
	1
		to: currentAddress - 1
		do:
			[:i | 
			newArray
				at: i
				put: (clauseArray at: i)].
	clauseArray := newArray! !

!LogicRepository methodsFor: 'private' stamp: 'sp 24/10/11 15:26'!
importClause: aClause
	"import a clause to this database, this happens when a repositoryvariable is bound to this repository and the other repository contains callclauses"

	
	| pred aCollection address multi |
	pred := aClause predicate.
	multi := aClause multiplicity.
	address := self storeAddressForPredicate: pred.
	aCollection := self clausesAt: address multiplicity: multi.
	aCollection add: aClause.
	self clausesAt: address multiplicity: multi put: aCollection! !

!LogicRepository methodsFor: 'initialize' stamp: 'sp 24/10/11 15:26'!
initialize
	"Called by the #new class method and initializes all necessary variables"

	
	maxAddress := 500.
	clauseArray := Array new: maxAddress.
	addressForPredicate := Dictionary new.
	currentAddress := 0.
	fromLayers := OrderedCollection new.
	repositoryVariables := Dictionary new.
	exports := Dictionary new.
	self loadNatives.
	storageChanged := false.
	name := 'unnamed repository'.
	tabledRules := OrderedCollection new.
	additionalClauses := OrderedCollection new! !

!LogicRepository methodsFor: 'access' stamp: 'sp 24/10/11 15:26'!
layers
	"returns the layers that were added to this repository using the #addLayer: method"

	
	^fromLayers! !

!LogicRepository methodsFor: 'add/del - from storage' stamp: 'sp 24/10/11 15:26'!
loadLayer: aLayer
	
	(fromLayers includes: aLayer) ifFalse: [fromLayers add: aLayer].
	Transcript
		show: 'Loading Layer: ' , aLayer asString;
		cr.
	storageChanged := true! !

!LogicRepository methodsFor: 'natives' stamp: 'AndyKellens 11/2/2011 09:53'!
loadNatives
	
	| natives |
	"The repository automatically contains a set of native predicates that can be used in Soul programs"

	natives := OrderedCollection new.
	natives
		add:
			(self
				makeNative: 'writeToProlog(?x)'
				if: SoulFactory current makeWriteToPrologClause).
	natives
		add:
			(self
				makeNative: 'getEnv(?var,?idx)'
				if: SoulFactory current makeGetEnvClause).
	natives
		add:
			(self
				makeNative: 'getRepository(?var)'
				if: SoulFactory current makeGetRepositoryClause).
	natives
		add:
			(self
				makeNative: 'setResolver(?block,?args)'
				if: SoulFactory current makeSetResolverClause).
	natives
		add:
			(self
				makeNative: 'not@(?clauses)'
				if: SoulFactory current makeNotClause).
	natives
		add:
			(self
				makeNative: 'or@(?clauses)'
				if: SoulFactory current makeOrClause).
	natives
		add:
			(self
				makeNative: 'and@(?clauses)'
				if: SoulFactory current makeAndClause).
	natives
		add:
			(self
				makeNative: 'envLookup(&term,?termApp)'
				if: SoulFactory current makeEnvLookupClause).
	natives
		add:
			(self
				makeNative: 'clause(&head,&body,?clause)'
				if: SoulFactory current makeClauseClause).
	self addProgram: (ProgramSequence clauses: natives)! !

!LogicRepository methodsFor: 'repository composition' stamp: 'sp 24/10/11 15:26'!
lookupBindingFor: aVariable
	"Lookup the binding for a repository variable"

	
	^self lookupBindingForName: aVariable name! !

!LogicRepository methodsFor: 'repository composition' stamp: 'sp 24/10/11 15:26'!
lookupBindingFor: aVariable ifAbsent: aBlock
	"Lookup the binding for a repository variable"

	
	^self
		lookupBindingForName: aVariable name
		ifAbsent: aBlock! !

!LogicRepository methodsFor: 'repository composition' stamp: 'sp 24/10/11 15:26'!
lookupBindingForName: varName
	"Lookup the binding for a repository variable"

	
	^self
		lookupBindingForName: varName
		ifAbsent: [self error: varName , ' does not exist!!']! !

!LogicRepository methodsFor: 'repository composition' stamp: 'sp 24/10/11 15:26'!
lookupBindingForName: varName ifAbsent: aBlock
	"Lookup the binding for a repository variable"

	
	^repositoryVariables at: varName ifAbsent: aBlock! !

!LogicRepository methodsFor: 'private' stamp: 'AndyKellens 11/2/2011 09:53'!
makeNative: aString if: aClauseClass
	
	| conclusion resultClause temp |
	"Production of native clauses"

	conclusion := (SoulFactory current makeParser
		parse: aString
		onError:
			[:text :pos | self error: 'Parser error: ' , text , 'at: ' , pos asString])
		clauses first conclusion.
	temp := OrderedCollection new.
	temp add: (aClauseClass newWith: conclusion termSequence).
	resultClause := SoulFactory current makeRule
		condition:
			(SoulFactory current makeTermSequence terms: temp containsCut: false)
		conclusion: conclusion.
	^resultClause! !

!LogicRepository methodsFor: 'access' stamp: 'sp 24/10/11 15:26'!
name
	
	^name! !

!LogicRepository methodsFor: 'access' stamp: 'sp 24/10/11 15:26'!
name: aName
	
	name := aName! !

!LogicRepository methodsFor: 'private' stamp: 'sp 24/10/11 15:26'!
nextAddress
	"returns the next free address in the repository"

	
	currentAddress := currentAddress + 1.
	currentAddress > maxAddress ifTrue: [self expandStorage: 500].
	^currentAddress! !

!LogicRepository methodsFor: 'access' stamp: 'sp 24/10/11 15:26'!
normalAddressLookup: anAddress multiplicity: aNumber
	"Do a lookup of a normal predicate using the address of the predicate and the required multiplicity. This method returns both the clauses whose conclusion matches the 	predicate / multiplicity exactly, as well as the 'variable multiplicity' predicates in the repository"

	
	^(self clausesAt: anAddress multiplicity: aNumber)
		, (self clausesAt: anAddress multiplicity: -1)! !

!LogicRepository methodsFor: 'natives' stamp: 'sp 24/10/11 15:26'!
reLoadNatives
	
	self retractNatives.
	self loadNatives! !

!LogicRepository methodsFor: 'add/del - runtime' stamp: 'sp 24/10/11 15:26'!
removeClause: aClause
	
	self removeClause: aClause cleanupGenerated: false! !

!LogicRepository methodsFor: 'add/del - runtime' stamp: 'sp 24/10/11 15:26'!
removeClause: aClause cleanupGenerated: aBoolean
	
	| pred multi address aCollection |
	"remove a clause from the database, the argument must be the same clause object as the one in the repository"

	pred := aClause predicate.
	multi := aClause multiplicity.
	address := self addressForPredicate: pred.
	aCollection := self clausesAt: address multiplicity: multi.
	aCollection
		remove: aClause
		ifAbsent: [self error: 'unknown error, tabled rules probably'].
	aBoolean
		ifTrue: [aClause isTabledRule ifTrue: [self removeTabledRule: aClause]]! !

!LogicRepository methodsFor: 'private' stamp: 'sp 24/10/11 15:26'!
removeGeneratedRules
	
	tabledRules
		do:
			[:aTabledRule | 
			| generated |
			generated := aTabledRule generatedRules.
			generated do: [:aRule | self removeClause: aRule].	"not sure if they are re-used or re-generated, so clearing to be safe"

			aTabledRule generatedRules: OrderedCollection new].
	tabledRules := OrderedCollection new! !

!LogicRepository methodsFor: 'add/del - from storage' stamp: 'sp 24/10/11 15:26'!
removeLayer: aClass
	"Remove the layer from this repository"

	
	fromLayers remove: aClass.
	storageChanged := true! !

!LogicRepository methodsFor: 'private' stamp: 'sp 24/10/11 15:26'!
removeTabledRule: aTabledRule
	
	aTabledRule generatedRules
		do: [:generated | self removeClause: generated cleanupGenerated: true].
	tabledRules
		remove: aTabledRule
		ifAbsent: [^self halt]! !

!LogicRepository methodsFor: 'private' stamp: 'sp 24/10/11 15:26'!
repositoryBindings
	
	^repositoryVariables! !

!LogicRepository methodsFor: 'add/del - runtime' stamp: 'sp 24/10/11 15:26'!
retract: aPredicateKey
	
	| pred multi |
	"Retract all clauses with a particular predicatekey = predicate / multiplicity"

	pred := LogicStorage predicateOfKey: aPredicateKey.
	multi := LogicStorage multiplicityOfKey: aPredicateKey.
	^self retract: pred multiplicity: multi! !

!LogicRepository methodsFor: 'add/del - runtime' stamp: 'sp 24/10/11 15:26'!
retract: aPredicate multiplicity: multi
	
	| address clauses |
	"Retract all clauses with a particular predicate and multiplicity"

	address := self addressForPredicate: aPredicate.
	clauses := self clausesAt: address multiplicity: multi.
	clauses isEmpty
		ifFalse:
			[self
				clausesAt: address
				multiplicity: multi
				put: (clauses allButFirst: 1)].
	^true! !

!LogicRepository methodsFor: 'add/del - runtime' stamp: 'sp 24/10/11 15:26'!
retractAll: aPredicateKey
	
	| pred multi |
	"Retract all clauses with a particular predicatekey = predicate / multiplicity"

	pred := LogicStorage predicateOfKey: aPredicateKey.
	multi := LogicStorage multiplicityOfKey: aPredicateKey.
	^self retractAll: pred multiplicity: multi! !

!LogicRepository methodsFor: 'add/del - runtime' stamp: 'sp 24/10/11 15:26'!
retractAll: aPredicate multiplicity: multi
	
	| address |
	"Retract all clauses with a particular predicate and multiplicity"

	address := self addressForPredicate: aPredicate.
	self
		clausesAt: address
		multiplicity: multi
		put: OrderedCollection new.
	^true! !

!LogicRepository methodsFor: 'natives' stamp: 'sp 24/10/11 15:26'!
retractNatives
	
	self
		clausesDo:
			[:clause | 
			| conditions |
			clause notNil
				ifTrue:
					["I encountered nil-clauses, this does not seem correct."

					conditions := clause condition terms.
					(conditions size >= 1 and: [conditions first isKindOf: NativeClause])
						ifTrue: [self removeClause: clause]]]! !

!LogicRepository methodsFor: 'private' stamp: 'sp 24/10/11 15:26'!
storeAddressForPredicate: predicate
	"sets and returns the address for this predicate"

	
	^addressForPredicate
		at: predicate
		ifAbsentPut: [self nextAddress]! !

!LogicRepository methodsFor: 'access' stamp: 'sp 24/10/11 15:26'!
tabledRules
	
	^tabledRules! !

!LogicRepository methodsFor: 'prepare for use' stamp: 'sp 24/10/11 15:26'!
transformTabledRules
	
	| newRules |
	"we generate the new rules"

	newRules := TabledRule transformTabledRules: tabledRules withCallGraph:
			callGraph.	"we remove the old tabled rules as they are not needed in the resolution. All the transformed rules are in newRules"

	tabledRules
		do:
			[:aTabledRule | self removeClause: aTabledRule cleanupGenerated: false].	"we add the newly generated rules to the repo"

	newRules do: [:aRule | self addClause: aRule].	"should probably be in seperate method"

	SLGCall initializeQueries! !

!LogicRepository methodsFor: 'update mechanism' stamp: 'AndyKellens 11/2/2011 09:53'!
update: anAspect with: aLayerClass
	"set the flag that this repository is no longer up-to-date"

	"check if the loaded layers have not been expanded with sublayers"

	
	anAspect == #addLayer
		ifTrue:
			[(fromLayers includes: aLayerClass superclass)
				ifTrue:
					[self loadLayer: aLayerClass.
					storageChanged := true]].
	anAspect == #removeLayer
		ifTrue:
			[(fromLayers includes: aLayerClass)
				ifTrue:
					[self removeLayer: aLayerClass.
					storageChanged := true]].
	anAspect == #changePredicate & (fromLayers includes: aLayerClass)
		ifTrue: [storageChanged := true].
	anAspect == #factoryChanged
		ifTrue: [storageChanged := (compilationFactory = SoulFactory current) not]! !

!LogicRepository methodsFor: 'access' stamp: 'sp 24/10/11 15:26'!
varArgsAddressLookup: anAddress
	"Do a lookup of a 'variable multiplicity' predicate using the address of the predicate. This method returns all clauses whose conclusion matches the predicate"

	
	^self allClausesAt: anAddress! !

!LogicRepository methodsFor: 'repository composition' stamp: 'sp 24/10/11 15:26'!
variables
	
	| visitor |
	visitor := RepositoryVariableAccVisitor new.
	self compile.
	1
		to: currentAddress
		do:
			[:index | 
			(clauseArray at: index)
				do:
					[:cl | cl == nil ifFalse: [cl do: [:clause | clause accept: visitor]]]].
	^visitor variables! !
LogicStorage initialize!
LogicRepository initialize!
